from astoptimizer import UNSET
from astoptimizer.compatibility import (
    INT_TYPES, STR_TYPES, UNICODE_TYPE,
    is_immutable)
import sys

ANY_TYPE = object()

DEFAULT_FEATURES = ('builtins', 'math')

# Don't optimize operations on non-BMP strings because the
# result is different on narrow and wide builds
def optimize_unicode(config, text):
    if 'pythonenv' in config.features:
        return True
    if sys.version_info >= (3,3):
        return True
    if sys.maxunicode == 0x10ffff:
        return all(ord(char) <= 0xffff for char in text)
    else:
        return all(not 0xd800 <= ord(char) <= 0xdfff for char in text)

class BaseFunction:
    def __init__(self, narg, *arg_types, **kw):
        """
        """
        if isinstance(narg, (tuple, list)):
            self.min_narg, self.max_narg = narg
        elif isinstance(narg, INT_TYPES):
            self.min_narg = narg
            self.max_narg = narg
        else:
            raise TypeError("expect tuple or int, got %s" % type(narg).__name__)
        self.arg_types = arg_types
        # def check_args(config, args, result):
        #    return True
        self.check_result_callback = kw.pop("check_result", None)
        if kw:
            raise ValueError("unknown arguments: %s" % ', '.join(kw.keys()))

    def _check_args(self, config, args):
        if not(self.min_narg <= len(args) <= self.max_narg):
            return False
        for index, arg in enumerate(args):
            types = self.arg_types[index]
            if types is ANY_TYPE:
                continue
            if not isinstance(arg, types):
                return False
        return True

    def check_result(self, config, result):
        if not config.check_result(result):
            return False
        if self.check_result_callback is not None:
            if not self.check_result_callback(config, result):
                return False
        return True

    def _call(self, config, func, args):
        try:
            result = func(*args)
        except OverflowError:
            return UNSET
        if result is UNSET:
            return UNSET
        if not self.check_result(config, result):
            return UNSET
        return result


class Function(BaseFunction):
    def __init__(self, func, narg, *arg_types, **kw):
        # def check_args(config, args):
        #    if ...:
        #        # can be done at compile time
        #        return True
        #    else:
        #        # must be done at runtime
        #        return False
        self.check_args_callback = kw.pop("check_args", None)
        BaseFunction.__init__(self, narg, *arg_types, **kw)
        self.func = func

    def call(self, config, args):
        if not self._check_args(config, args):
            return UNSET
        if self.check_args_callback is not None:
            if not self.check_args_callback(config, args):
                return UNSET
        return self._call(config, self.func, args)


class Method(BaseFunction):
    def __init__(self, obj_type, method, narg, *arg_types, **kw):
        # def check_args(config, obj, args):
        #    if ...:
        #        # can be done at compile time
        #        return True
        #    else:
        #        # must be done at runtime
        #        return False
        self.check_args_callback = kw.pop("check_args", None)
        BaseFunction.__init__(self, narg, *arg_types, **kw)
        self.obj_type = obj_type
        self.method = method

    def call(self, config, obj, args):
        if not self._check_args(config, args):
            return UNSET
        if self.check_args_callback is not None:
            if not self.check_args_callback(config, obj, args):
                return UNSET
        func = getattr(obj, self.method)
        return self._call(config, func, args)


class Config:
    def __init__(self, *features):
        self.features = set(DEFAULT_FEATURES + features)

        # Hardcoded constants
        self._constants = {
            'True': True,
            'False': False,
            'None': None,
        }

        # List of functions that can be called at compile time (no border
        # effect), dictionnary of name => (check_callback, func) with
        # check_callback returning False if the call cannot be done:
        #
        # def check_callback(config, args):
        #    if ...:
        #        # can be done at compile time
        #        return True
        #    else:
        #        # must be done at runtime
        #        return False
        self.functions = {}

        # List of methods that can be called at compile time (no border
        # effect), dictionnary of type => {attr_name => check_callback}
        # with check_callback returning False if the call cannot be done:
        #
        # def check_callback(config, obj, args):
        #    if ...:
        #        # can be done at compile time
        #        return True
        #    else:
        #        # must be done at runtime
        #        return False
        self.methods = {}

        # Minimum and maximum value of integers (int and long types).
        # Arbitrary limit to limit memory footprint and disk space
        # (max 128 bits per number).
        self.min_int = -2 ** 128
        self.max_int = 2 ** 128 - 1

        # Maximum length of a string in characters (bytes type) or bytes (str type)
        self.max_string_length = 4096

        # Maximum number of items of a tuple or a frozenset
        self.max_tuple_length = 9

        # The maximum size is at least 2^31-1 according to Python 2.7
        # documentation
        self.max_size = 2**31 - 1

        for feature in self.features:
            modname = "config_%s" % feature
            mod = __import__("astoptimizer.%s" % modname)
            mod = getattr(mod, modname)
            mod.setup_config(self)

    def add_constant(self, name, value):
        if not is_immutable(value):
            raise TypeError("Unsupport constant type: %r (%s)"
                            % (value, type(value).__name__))
        if name in self._constants:
            raise ValueError("Duplicate constant: %r" % (name,))
        self._constants[name] = value

    def get_constant(self, name):
        return self._constants.get(name, UNSET)

    def add_func(self, name, func):
        if name in self.functions:
            raise ValueError("Duplicate function: %r" % (name,))
        self.functions[name] = func

    def add_method(self, method):
        obj_type = method.obj_type
        attr = method.method
        if obj_type in self.methods:
            if attr in self.methods[obj_type]:
                raise ValueError("Duplicate method %s.%r" % (obj_type, attr))
            self.methods[obj_type][attr] = method
        else:
            self.methods[obj_type] = {attr: method}

    def check_result(self, result):
        if result is None:
            return True
        if isinstance(result, INT_TYPES):
            return (self.min_int <= result <= self.max_int)
        if isinstance(result, STR_TYPES):
            if len(result) > self.max_string_length:
                return False
            if (isinstance(result, UNICODE_TYPE)
            and not optimize_unicode(self, result)):
                return False
            return True
        if isinstance(result, (float, complex)):
            return True
        if isinstance(result, (tuple, frozenset)):
            if len(result) > self.max_tuple_length:
                return False
            return all(self.check_result(arg) for arg in result)
        return False

