"""
Enable optimizations on the builtins functions (ex: len()) and types (ex:
unicode.encode()).
"""

from astoptimizer.compatibility import (
    PYTHON3,
    INT_TYPES, FLOAT_TYPES, COMPLEX_TYPES,
    NATIVE_STR, BYTES_TYPE, UNICODE_TYPE, STR_TYPES,
    IMMUTABLE_ITERABLE_TYPES, IMMUTABLE_TYPES)
from astoptimizer.config import Function, Method, ANY_TYPE
import codecs
import sys

def check_ord_args(config, args):
    arg = args[0]
    return (len(arg) == 1)

def check_max_args(config, args):
    arg = args[0]
    if isinstance(arg, STR_TYPES):
        return True
    return all(isinstance(item, COMPLEX_TYPES) for item in arg)

def check_sum_args(config, args):
    arg = args[0]
    if not all(isinstance(item, COMPLEX_TYPES) for item in arg):
        return False
    if len(args) == 2:
        if not isinstance(args[1], COMPLEX_TYPES):
            return False
    return True

def check_tuple_args(config, args):
    if len(args) == 0:
        return True
    arg = args[0]
    if not isinstance(arg, IMMUTABLE_ITERABLE_TYPES):
        return False
    if len(arg) > config.max_tuple_length:
        return False
    return all(isinstance(item, IMMUTABLE_TYPES) for item in arg)

def check_int(config, args):
    arg = args[0]
    if isinstance(arg, STR_TYPES):
        try:
            int(arg)
        except ValueError:
            return False
    return True

def check_float(config, args):
    arg = args[0]
    if isinstance(arg, STR_TYPES):
        try:
            float(arg)
        except ValueError:
            return False
    return True

def check_len(config, result):
    return (result <= config.max_size)

if not PYTHON3:
    def check_unicode_args(config, args):
        arg = args[0]
        if not isinstance(arg, BYTES_TYPE):
            return True
        return all(ord(char) <= 127 for char in arg)

def check_encoding(encoding):
    try:
        encoding = codecs.lookup(encoding).name
    except LookupError:
        return False
    if encoding not in ('ascii', 'iso8859-1', 'utf-8'):
        return False
    return True

def check_str_decode_args(config, obj, args):
    if not check_encoding(args[0]):
        return False
    if len(args) == 2:
        errors = args[1]
        if errors not in ('strict', 'replace', 'ignore'):
            return False
    return True

def check_str_encode_args(config, obj, args):
    if not check_encoding(args[0]):
        return False
    if len(args) == 2:
        errors = args[1]
        if errors not in ('strict', 'replace', 'ignore', 'backslashreplace'):
            return False
    return True

def check_index_args(config, obj, args):
    return (args[0] in obj)

def check_pow(config, args):
    if len(args) >= 3:
        if not all(isinstance(arg, INT_TYPES) for arg in args[:2]):
            return False

    num = args[0]
    exp = args[1]
    if exp < 1.0 and exp != 0.0 and num < 0:
        # pow(-25, 0.5) raises a ValueError
        return False
    if len(args) >= 3:
        mod = args[2]
        if mod == 0:
            # pow(2, 1024, 0) raises a ValueError('pow() 3rd argument cannot be 0')
            return False
    return True

def check_chr(config, args):
    code = args[0]
    return 0 <= code <= 0xff

def check_unichr(config, args):
    code = args[0]
    if PYTHON3:
        return 0 <= code <= 0x10ffff
    else:
        return 0 <= code <= sys.maxunicode

def setup_config(config):
    # pure builtin functions
    config.add_func('abs', Function(abs, 1, COMPLEX_TYPES))
    config.add_func('bool', Function(bool, 1, FLOAT_TYPES + STR_TYPES))
    if PYTHON3:
        config.add_func('chr', Function(chr, 1, INT_TYPES, check_args=check_unichr))
    else:
        config.add_func('chr', Function(chr, 1, INT_TYPES, check_args=check_chr))
    config.add_func('divmod', Function(divmod, 2, FLOAT_TYPES, FLOAT_TYPES))
    config.add_func('float', Function(float, 1, FLOAT_TYPES + STR_TYPES, check_args=check_float))
    config.add_func('int', Function(int, 1, FLOAT_TYPES + STR_TYPES, check_args=check_int))
    config.add_func('len', Function(len, 1, IMMUTABLE_ITERABLE_TYPES, check_result=check_len))
    config.add_func('oct', Function(oct, 1, INT_TYPES))
    config.add_func('ord', Function(ord, 1, STR_TYPES, check_args=check_ord_args))
    config.add_func('min', Function(min, 1, STR_TYPES + (tuple, frozenset), check_args=check_max_args))
    config.add_func('max', Function(max, 1, STR_TYPES + (tuple, frozenset), check_args=check_max_args))
    config.add_func('pow', Function(pow, (2, 3), FLOAT_TYPES, FLOAT_TYPES, FLOAT_TYPES, check_args=check_pow))
    config.add_func('repr', Function(repr, 1, STR_TYPES + COMPLEX_TYPES))
    config.add_func('round', Function(round, (1, 2), FLOAT_TYPES, INT_TYPES))
    config.add_func('str', Function(str, 1, STR_TYPES + COMPLEX_TYPES))
    config.add_func('sum', Function(sum, (1, 2), (tuple, frozenset), COMPLEX_TYPES, check_args=check_sum_args))
    config.add_func('tuple', Function(tuple, (0, 1), IMMUTABLE_ITERABLE_TYPES, check_args=check_tuple_args))

    if sys.version_info >= (2,6):
        config.add_func('bin', Function(bin, 1, INT_TYPES))
    if not PYTHON3:
        config.add_func('long', Function(long, 1, FLOAT_TYPES))
        config.add_func('unichr', Function(unichr, 1, INT_TYPES, check_args=check_unichr))
        config.add_func('unicode', Function(unicode, 1, STR_TYPES + COMPLEX_TYPES, check_args=check_unicode_args))

    # bytes, str methods
    config.add_method(Method(BYTES_TYPE, 'decode', (1, 2), NATIVE_STR, NATIVE_STR, check_args=check_str_decode_args))
    config.add_method(Method(UNICODE_TYPE, 'encode', (1, 2), NATIVE_STR, NATIVE_STR, check_args=check_str_encode_args))
    for str_type in STR_TYPES:
        config.add_method(Method(str_type, 'count', 1, str_type))
        config.add_method(Method(str_type, 'endswith', 1, str_type))
        config.add_method(Method(str_type, 'find', 1, str_type))
        config.add_method(Method(str_type, 'index', 1, str_type, check_args=check_index_args))
        config.add_method(Method(str_type, 'startswith', 1, str_type))
    if sys.version_info >= (2, 6):
        config.add_method(Method(tuple, 'count', 1, ANY_TYPE))
        config.add_method(Method(tuple, 'index', 1, ANY_TYPE, check_args=check_index_args))

    # int, float methods
    config.add_method(Method(int, 'bit_length', 0))
    if not PYTHON3:
        config.add_method(Method(long, 'bit_length', 0))

    if sys.version_info >= (2,6):
        config.add_method(Method(float, 'is_integer', 0))
        config.add_method(Method(float, 'as_integer_ratio', 0))
        config.add_method(Method(float, 'hex', 0))
        config.add_func('float.fromhex', Function(float.fromhex, 1, NATIVE_STR))

