import operator
from astoptimizer import UNSET
from astoptimizer.config import Config, optimize_unicode
from astoptimizer.compatibility import (
    ast,
    PYTHON3,
    is_bytes_ascii, is_singleton,
    INT_TYPES, FLOAT_TYPES, COMPLEX_TYPES,
    NATIVE_STR, BYTES_TYPE, UNICODE_TYPE, STR_TYPES,
    IMMUTABLE_ITERABLE_TYPES)
import sys

PYTHON26 = (sys.version_info >= (2, 6))

# FIXME: support scope
ENABLE_VARIABLES = False

def operator_not_in(data, item):
    return item not in data

EVAL_COMPARE = {
    # a == b, a != b
    ast.Eq: operator.eq,
    ast.NotEq: operator.ne,
    # a < b, a <= b
    ast.Lt: operator.lt,
    ast.LtE: operator.le,
    # a > b, a >= b
    ast.Gt: operator.gt,
    ast.GtE: operator.ge,
    # a in b, a not in b
    ast.In: operator.contains,
    ast.NotIn: operator_not_in,
    # a is b, a is not b
    ast.Is: operator.is_,
    ast.IsNot: operator.is_not,
}

EVAL_UNARYOP = {
    # not a, ~a, +a, -a
    ast.Not: operator.not_,
    ast.Invert: operator.invert,
    ast.UAdd: operator.pos,
    ast.USub: operator.neg,
}

EVAL_BINOP = {
    # a + b, a - b, a * b
    ast.Add: operator.add,
    ast.Sub: operator.sub,
    ast.Mult: operator.mul,
    # see binop(): floordiv() may be used for int/int on Python 2
    ast.Div: operator.truediv,
    ast.FloorDiv: operator.floordiv,
    ast.Mod: operator.mod,
    # a ** b
    ast.Pow: operator.pow,
    # a << b, a >> b
    ast.LShift: operator.lshift,
    ast.RShift: operator.rshift,
    # a & b, a | b, a ^ b
    ast.BitAnd: operator.and_,
    ast.BitOr: operator.or_,
    ast.BitXor: operator.xor,
}

FLOAT_BINOPS = (
    ast.Add, ast.Sub,
    ast.Mult, ast.Div, ast.FloorDiv, ast.Mod,
    ast.Pow)

DIVIDE_BINOPS = (ast.Div, ast.FloorDiv, ast.Mod)
COMPLEX_BINOPS = (
    ast.Add, ast.Sub,
    ast.Mult, ast.Div)
if not PYTHON3:
    COMPLEX_BINOPS += (ast.FloorDiv, ast.Mod)

TUPLE_BINOPS = (ast.Add,)
FROZENSET_BINOPS = (ast.BitAnd, ast.BitOr, ast.BitXor, ast.Sub)

EVAL_BOOLOP = {
    # a and b, a or b
    ast.And: operator.__and__,
    ast.Or: operator.__or__,
}

class Namespace:
    def __init__(self):
        self._removed_aliases = set()
        self.qualnames = {}
        self.values = {}
        self._unassigned = set()
        self.func_args = set()

    def add_alias(self, qualname, name):
        """
        Register name as an alias to qualname.

        If name is already registered, unregister the previous
        alias. If the name was already removed, don't register
        the new name.
        """
        if name in self._removed_aliases:
            return
        if name in self.qualnames:
            self._removed_aliases.add(name)
            del self.qualnames[name]
            return
        self.qualnames[name] = qualname

    def get_qualname(self, name):
        return self.qualnames.get(name, name)

    def assign(self, qualname, value):
        if qualname in self._unassigned:
            return
        self.values[qualname] = value

    def unassign(self, qualname):
        self._unassigned.add(qualname)
        self.values.pop(qualname, None)

    def get_var(self, name):
        qualname = self.get_qualname(name)
        return self.values.get(qualname, UNSET)


class Optimizer(ast.NodeTransformer):
    def __init__(self, config=None):
        ast.NodeTransformer.__init__(self)
        if config is None:
            config = Config()
        self.config = config
        self.namespace = None

    def optimize(self, tree):
        self.namespace = Namespace()
        return self.visit(tree)

    def fix_node(self, node, new_node):
        # FIXME: use the right lineno / col_offset
        new_node.lineno = 1
        new_node.col_offset = 1
        ast.fix_missing_locations(new_node)
        return ast.copy_location(new_node, node)

    def get_constant(self, node):
        if node is None:
            return None
        if isinstance(node, ast.Num):
            return node.n
        if isinstance(node, ast.Str):
            text = node.s
            if (isinstance(text, UNICODE_TYPE)
            and not optimize_unicode(self.config, text)):
                return UNSET
            return text
        if PYTHON3:
            if isinstance(node, ast.Bytes):
                return node.s
        if isinstance(node, ast.Name):
            name = node.id
            qualname = self.namespace.get_qualname(name)
            constant = self.config.get_constant(qualname)
            if constant is not UNSET:
                return constant
            return self.namespace.get_var(name)
        if isinstance(node, ast.Tuple):
            elts = node.elts
            if len(elts) > self.config.max_tuple_length:
                return UNSET
            constants = []
            for elt in elts:
                constant = self.get_constant(elt)
                if constant is UNSET:
                    return UNSET
                constants.append(constant)
            return tuple(constants)
        if (isinstance(node, ast.Call)
        and isinstance(node.func, ast.Name)
        and node.func.id == 'frozenset'):
            if len(node.args) == 1:
                arg = self.get_constant(node.args[0])
                if arg is UNSET:
                    return UNSET
                if not isinstance(arg, IMMUTABLE_ITERABLE_TYPES):
                    return UNSET
                if len(arg) > self.config.max_tuple_length:
                    return UNSET
                return frozenset(arg)
            elif len(node.args) == 0:
                return frozenset()
        return UNSET

    def get_constant_list(self, nodes):
        constants = []
        for node in nodes:
            constant = self.get_constant(node)
            if constant is UNSET:
                return UNSET
            constants.append(constant)
        return constants

    if PYTHON26:
        def new_constant(self, node, value):
            if isinstance(value, bool):
                name = "True" if value else "False"
                new_node = ast.Name(id=name, ctx=ast.Load())
            elif isinstance(value, COMPLEX_TYPES):
                new_node = ast.Num(n=value)
            elif isinstance(value, UNICODE_TYPE):
                if PYTHON3:
                    new_node = ast.Str(s=value)
                else:
                    new_node = ast.Str(s=value)
            elif isinstance(value, BYTES_TYPE):
                if PYTHON3:
                    new_node = ast.Bytes(s=value)
                else:
                    new_node = ast.Str(s=value)
            elif value is None:
                new_node = ast.Name(id="None", ctx=ast.Load())
            elif isinstance(value, tuple):
                elts = [self.new_constant(node, elt) for elt in value]
                new_node = ast.Tuple(elts=elts, ctx=ast.Load())
            elif isinstance(value, frozenset):
                elts = [self.new_constant(node, elt) for elt in value]
                arg = ast.Tuple(elts=elts, ctx=ast.Load())
                func = ast.Name(id='frozenset', ctx=ast.Load())
                new_node = ast.Call(func, [arg], [], None, None)
            else:
                raise NotImplementedError("unable to create an AST object for constant: %r" % (value,))
            return self.fix_node(node, new_node)
    else:
        # Implementation for Python 2.5
        def new_constant(self, node, value):
            if isinstance(value, bool):
                name = "True" if value else "False"
                new_node = ast.Name()
                new_node.id = name
                new_node.ctx = ast.Load()
            elif isinstance(value, COMPLEX_TYPES):
                new_node = ast.Num()
                new_node.n = value
            elif isinstance(value, (UNICODE_TYPE, BYTES_TYPE)):
                new_node = ast.Str()
                new_node.s = value
            elif value is None:
                new_node = ast.Name()
                new_node.id = "None"
                new_node.ctx = ast.Load()
            elif isinstance(value, tuple):
                elts = [self.new_constant(node, elt) for elt in value]
                new_node = ast.Tuple()
                new_node.elts = elts
                new_node.ctx = ast.Load()
            elif isinstance(value, frozenset):
                elts = [self.new_constant(node, elt) for elt in value]
                arg = ast.Tuple()
                arg.elts = elts
                arg.ctx = ast.Load()

                func = ast.Name()
                func.id = 'frozenset'
                func.ctx = ast.Load()

                new_node = ast.Call()
                new_node.func = func
                new_node.args = [arg]
                new_node.keywords = []
                new_node.starargs = None
                new_node.kwargs  = None
            else:
                raise NotImplementedError("unable to create an AST object for constant: %r" % (value,))
            return self.fix_node(node, new_node)

    def visit(self, node):
        method = 'fullvisit_' + node.__class__.__name__
        visitor = getattr(self, method, None)
        if visitor is not None:
            new_node = visitor(node)
            if new_node is not None:
                return new_node
            else:
                return node

        new_node = self.generic_visit(node)

        method = 'visit_' + node.__class__.__name__
        visitor = getattr(self, method, UNSET)
        if visitor is UNSET:
            return new_node

        optimized_node = visitor(new_node)
        if optimized_node is not None:
            return optimized_node
        else:
            return new_node

    def visit_Name(self, node):
        constant = self.get_constant(node)
        if constant is UNSET:
            return
        return self.new_constant(node, constant)

    def visit_Attribute(self, node):
        if isinstance(node.value, ast.Name):
            fullname = "%s.%s" % (node.value.id, node.attr)
            constant = self.config.get_constant(fullname)
            if constant is UNSET:
                return
            return self.new_constant(node, constant)
        return

    def string_format(self, node, format, arg):
        try:
            result = format % arg
        except (TypeError, ValueError, UnicodeError):
            # format string or arguments are invalid
            return UNSET
        if (isinstance(format, BYTES_TYPE)
        and isinstance(result, UNICODE_TYPE)
        and not is_bytes_ascii(format)):
            # b'x\xe9=%s' % u'abc' gives a different result
            # depending on the default encoding
            return UNSET
        if not self.config.check_result(result):
            return UNSET
        return self.new_constant(node, result)

    def check_binop_cst(self, op, left, right):
        if (isinstance(left, COMPLEX_TYPES)
        and isinstance(right, COMPLEX_TYPES)):
            if (isinstance(op, DIVIDE_BINOPS)
            and not right):
                # division by zero
                return False

            if (isinstance(op, ast.Pow)
            and isinstance(left, FLOAT_TYPES)
            and isinstance(right, FLOAT_TYPES)):
                # -1.0 ** 0.5 raises a ValueError
                if right < 1.0 and right != 0.0 and left < 0:
                    return False

            if (isinstance(op, (ast.LShift, ast.RShift))
            and right < 0):
                # 1 << -3 and 1 >> -3 raise a ValueError
                return False

        if (isinstance(left, INT_TYPES)
        and isinstance(right, INT_TYPES)):
            return True

        if (isinstance(left, FLOAT_TYPES)
        and isinstance(right, FLOAT_TYPES)):
            return isinstance(op, FLOAT_BINOPS)

        if (isinstance(left, COMPLEX_TYPES)
        and isinstance(right, COMPLEX_TYPES)):
            return isinstance(op, COMPLEX_BINOPS)

        if isinstance(op, ast.Mult):
            if isinstance(right, INT_TYPES):
                # str * int
                if isinstance(left, STR_TYPES):
                    return (len(left) * right <= self.config.max_string_length)
                # tuple * int
                if isinstance(left, tuple):
                    return (len(left) * right <= self.config.max_tuple_length)

            if isinstance(left, INT_TYPES):
                # int * str
                if isinstance(right, STR_TYPES):
                    return (left * len(right) <= self.config.max_string_length)
                # int * tuple
                if isinstance(right, tuple):
                    return (left * len(right) <= self.config.max_tuple_length)

        if (isinstance(left, tuple)
        and isinstance(right, tuple)):
            return isinstance(op, TUPLE_BINOPS)

        if (isinstance(left, frozenset)
        and isinstance(right, frozenset)):
            return isinstance(op, FROZENSET_BINOPS)

        return False

    def binop(self, node, eval_binop, left, left_cst, right, right_cst):
        if left_cst is UNSET:
            return UNSET
        if right_cst is UNSET:
            return UNSET

        if (isinstance(node.op, ast.Mod)
        and isinstance(left_cst, STR_TYPES)):
            # str % args
            if PYTHON3 and isinstance(left_cst, BYTES_TYPE):
                return UNSET
            return self.string_format(node, left_cst, right_cst)

        if (not PYTHON3
        and isinstance(node.op, ast.Div)
        and (isinstance(left_cst, INT_TYPES)
             or isinstance(right_cst, INT_TYPES))):
            # Don't optimize a / b on Python 2, because it depends on the
            # -Q command line option, except if pythonenv feature is enable
            if 'pythonenv' not in self.config.features:
                return UNSET
            if sys.version_info >= (2,6):
                if not sys.flags.division_new:
                    eval_binop = operator.floordiv
            else:
                if (1 / 2) == 0:
                    eval_binop = operator.floordiv

        ok = self.check_binop_cst(node.op, left_cst, right_cst)
        if not ok:
            return UNSET

        result = eval_binop(left_cst, right_cst)
        if not self.config.check_result(result):
            return UNSET
        return self.new_constant(node, result)

    def visit_BinOp(self, node):
        eval_binop = EVAL_BINOP.get(node.op.__class__)
        if not eval_binop:
            return
        left = node.left
        left_cst = self.get_constant(left)
        right = node.right
        right_cst = self.get_constant(right)
        new_node = self.binop(node, eval_binop, left, left_cst, right, right_cst)
        if new_node is not UNSET:
            return new_node
        return node

    def visit_BoolOp(self, node):
        # FIXME: use EVAL_BOOLOP
        if not isinstance(node.op, (ast.And, ast.Or)):
            return
        if len(node.values) < 2:
            return
        values = node.values
        constants = []
        for value in values:
            constant = self.get_constant(value)
            if constant is UNSET:
                break
            constants.append(constant)
        else:
            # all arguments are constant
            result = constants[0]
            if isinstance(node.op, ast.And):
                for constant in constants[1:]:
                    result = result and constant
            else:
                for constant in constants[1:]:
                    result = result or constant
            return self.new_constant(node, result)

        result = values
        for index, constant in enumerate(constants):
            test = constant
            if isinstance(node.op, ast.Or):
                test = not constant
            if test:
                result = values[index+1:]
            else:
                return self.new_constant(node, constant)
        if len(result) > 1:
            if PYTHON26:
                new_node = ast.BoolOp(node.op, result)
            else:
                new_node = ast.BoolOp()
                new_node.op = node.op
                new_node.values = result
            return self.fix_node(node, new_node)
        else:
            return result[0]

    def visit_UnaryOp(self, node):
        eval_unaryop = EVAL_UNARYOP.get(node.op.__class__)
        if eval_unaryop is None:
            return
        operand = node.operand
        constant = self.get_constant(operand)
        if constant is UNSET:
            return
        result = eval_unaryop(constant)
        if not self.config.check_result(result):
            return
        return self.new_constant(node, result)

    def subscript_slice(self, node, value, lower, upper, step):
        lower_cst = self.get_constant(lower)
        if lower_cst is UNSET:
            return UNSET
        upper_cst = self.get_constant(upper)
        if upper_cst is UNSET:
            return UNSET
        step_cst = self.get_constant(step)
        if step_cst is UNSET:
            return UNSET
        myslice = slice(lower_cst, upper_cst, step_cst)
        return self.new_constant(node, value[myslice])

    def subscript_index(self, node, value, index):
        index_constant = self.get_constant(index)
        if index_constant is UNSET:
            return UNSET
        if index_constant >= 0:
            if index_constant >= len(value):
                return UNSET
        else:
            if (len(value) + index_constant) < 0:
                return UNSET
        result = value[index_constant]
        return self.new_constant(node, result)

    def visit_Subscript(self, node):
        value = node.value
        value_constant = self.get_constant(value)
        if value_constant is UNSET:
            return
        if isinstance(node.slice, ast.Slice):
            lower = node.slice.lower
            upper = node.slice.upper
            step = node.slice.step
            new_node = self.subscript_slice(node, value_constant, lower, upper, step)
            if new_node is not UNSET:
                return new_node
        elif isinstance(node.slice, ast.Index):
            index = node.slice.value
            new_node = self.subscript_index(node, value_constant, index)
            if new_node is not UNSET:
                return new_node

    def check_compare_in(self, item, data):
        if isinstance(data, (tuple, frozenset)):
            return True
        if (isinstance(data, STR_TYPES)
        and type(item) == type(data)):
            return True
        return False

    def check_compare(self, left, op, right):
        if isinstance(op, (ast.Is, ast.IsNot)):
            return (is_singleton(self.config, left)
                    and is_singleton(self.config, right))

        if (isinstance(left, COMPLEX_TYPES)
        and isinstance(right, COMPLEX_TYPES)):
            return True

        if (isinstance(left, STR_TYPES + (tuple, frozenset))
        and type(right) == type(left)):
            return True

        return False

    def compare(self, node, op, left_cst, comparators):
        eval_compare = EVAL_COMPARE.get(op.__class__)
        if eval_compare is None:
            return UNSET
        # FIXME: can it be longer than 1?
        assert len(comparators) == 1
        right = comparators[0]
        right_cst = self.get_constant(right)
        if right_cst is UNSET:
            return UNSET
        if isinstance(op, (ast.In, ast.NotIn)):
            if not self.check_compare_in(left_cst, right_cst):
                return UNSET
            result = eval_compare(right_cst, left_cst)
        else:
            if not self.check_compare(left_cst, op, right_cst):
                return UNSET
            result = eval_compare(left_cst, right_cst)
        if not self.config.check_result(result):
            return UNSET
        return self.new_constant(node, result)

    def visit_Compare(self, node):
        left = node.left
        left_cst = self.get_constant(left)
        comparators = node.comparators
        if left_cst is UNSET:
            return
        if len(node.ops) != 1:
            return
        op = node.ops[0]
        new_node = self.compare(node, op, left_cst, comparators)
        if new_node is UNSET:
            return
        return new_node

    def if_block(self, parent, nodes):
        nodes = nodes
        # FIXME: drop Pass?
        #nodes = [node for node in nodes if not isinstance(node, ast.Pass)]
        if len(nodes) == 1:
            nodes = nodes[0]
        elif not nodes:
            nodes = ast.Pass()
            nodes = self.fix_node(parent, nodes)
        return nodes

    def visit_If(self, node):
        new_test = node.test
        constant = self.get_constant(new_test)
        if constant is UNSET:
            return
        if constant:
            return self.if_block(node, node.body)
        else:
            return self.if_block(node, node.orelse)

    def visit_While(self, node):
        new_test = node.test
        constant = self.get_constant(new_test)
        if (constant is UNSET) or constant:
            return
        # while False: ...
        orelse = node.orelse
        if orelse:
            return orelse
        else:
            new_node = ast.Pass()
            return self.fix_node(node, new_node)

    def print_func(self, node, args):
        constants = self.get_constant_list(args)
        if constants is UNSET:
            return UNSET
        text = ' '.join(str(constant) for constant in constants)
        text = self.new_constant(node, text)
        func = ast.Name(id='print', ctx=ast.Load())
        new_node = ast.Call(func, [text], [], None, None)
        return self.fix_node(node, new_node)

    def call_func(self, node, name, args):
        if PYTHON3 and name == 'print':
            return self.print_func(node, args)
        if name not in self.config.functions:
            return UNSET
        func = self.config.functions[name]
        constants = self.get_constant_list(args)
        if constants is UNSET:
            return UNSET
        result = func.call(self.config, constants)
        if result is UNSET:
            return UNSET
        return self.new_constant(node, result)

    def call_method(self, node, func, args):
        obj_cst = self.get_constant(func.value)
        if obj_cst is UNSET:
            return UNSET
        attr = func.attr
        assert isinstance(attr, NATIVE_STR)

        obj_type = type(obj_cst)
        if obj_type not in self.config.methods:
            return UNSET
        methods = self.config.methods[obj_type]
        if attr not in methods:
            return UNSET
        method = methods[attr]

        constants = self.get_constant_list(args)
        if constants is UNSET:
            return UNSET
        result = method.call(self.config, obj_cst, constants)
        if result is UNSET:
            return UNSET
        return self.new_constant(node, result)

    def visit_Call(self, node):
        func = node.func
        args = node.args
        keywords = node.keywords
        starargs = node.starargs
        kwargs = node.kwargs
        # Don't support keywords, *args, **kw yet
        if (not keywords
        and not starargs
        and not kwargs):
            if isinstance(func, ast.Name):
                name = self.namespace.get_qualname(func.id)
                new_node = self.call_func(node, name, args)
                if new_node is not UNSET:
                    return new_node
            elif isinstance(func, ast.Attribute) and isinstance(func.value, ast.Name):
                name = self.namespace.get_qualname(func.value.id)
                name = "%s.%s" % (name, func.attr)
                new_node = self.call_func(node, name, args)
                if new_node is not UNSET:
                    return new_node
            elif isinstance(func, ast.Attribute):
                new_node = self.call_method(node, func, args)
                if new_node is not UNSET:
                    return new_node

    if not PYTHON3:
        def visit_Print(self, node):
            constants = self.get_constant_list(node.values)
            if constants is UNSET:
                return
            text = ' '.join(str(constant) for constant in constants)
            text = self.new_constant(node, text)
            if PYTHON26:
                new_node = ast.Print(node.dest, [text], node.nl)
            else:
                new_node = ast.Print()
                new_node.dest = node.dest
                new_node.values = [text]
                new_node.nl = node.nl
            return self.fix_node(node, new_node)

    def visit_ImportFrom(self, node):
        # from sys import platform: sys -> sys.platform
        # from sys import platform as plat: plat -> sys.platform
        module = node.module
        names = node.names
        for alias in names:
            assert isinstance(alias, ast.alias)
            qualname = "%s.%s" % (module, alias.name)
            if alias.asname:
                name = alias.asname
            else:
                name = alias.name
            self.namespace.add_alias(qualname, name)
        return node

    def visit_Import(self, node):
        # import math as M: M -> math
        for alias in node.names:
            assert isinstance(alias, ast.alias)
            if not alias.asname:
                continue
            self.namespace.add_alias(alias.name, alias.asname)
        return node

    if ENABLE_VARIABLES:
        def assign_multiple(self, names, values):
            if len(names) != len(values):
                return False
            values = self.get_constant_list(values)
            if values is UNSET:
                return False
            for name, value in zip(names, values):
                self.namespace.assign(name, value)
            return True

        def _get_assign_name(self, main_node, node, strict=True):
            if isinstance(node, ast.Name):
                return (node.id, True)
            elif isinstance(node, ast.Attribute):
                name = "%s.%s" % (self._get_assign_name(main_node, node.value), node.attr)
                return (name, True)
            elif isinstance(node, ast.Subscript):
                name, supported = self._get_assign_name(main_node, node.value)
                return (name, False)
            elif isinstance(node, ast.Call):
                return (None, False)
            elif strict:
                raise NotImplementedError("%s in %s" % (ast.dump(node), ast.dump(main_node)))
            else:
                return None

        def assign_single(self, name, supported, value):
            if name is None:
                return

            if not supported:
                self.namespace.unassign(name)
                return

            value = self.get_constant(value)
            if value is UNSET:
                self.namespace.unassign(name)
                return
            self.namespace.assign(name, value)

        def assign(self, node, target, value, assign_supported):
            result = self._get_assign_name(node, target, None)
            if result is not None:
                name, supported = result
                self.assign_single(name, assign_supported and supported, value)
                return

            if not isinstance(target, ast.Tuple):
                raise NotImplementedError(ast.dump(target))

            names = []
            assert len(target.elts) > 1
            supported = assign_supported
            for elt in target.elts:
                name, elt_supported = self._get_assign_name(node, elt)
                names.append(name)
                supported &= elt_supported

            if supported and isinstance(value, ast.Tuple):
                values = value.elts
                if not self.assign_multiple(names, values):
                    for name in names:
                        self.namespace.unassign(name)
            else:
                # Example: a, b = func()
                for name in names:
                    if name is None:
                        continue
                    self.namespace.unassign(name)

        def fullvisit_Assign(self, node):
            node.value = self.visit(node.value)
            for target in node.targets:
                self.assign(node, target, node.value, True)

        def fullvisit_AugAssign(self, node):
            node.value = self.visit(node.value)
            self.assign(node, node.target, node.value, False)

        def fullvisit_Delete(self, node):
            for target in node.targets:
                self.assign(node, target, None, False)
    else:
        def fullvisit_Assign(self, node):
            node.value = self.visit(node.value)
        def fullvisit_AugAssign(self, node):
            node.value = self.visit(node.value)
        def fullvisit_Delete(self, node):
            pass

