from __future__ import with_statement
import sys

import unittest
from astoptimizer import Config, parse_ast
if sys.version_info >= (2,6):
    from astoptimizer import compile_ast
from astoptimizer.compatibility import PYTHON3, u, b, ast, BYTES_TYPE, UNICODE_TYPE
from astoptimizer.optimizer import Namespace, Optimizer, ENABLE_VARIABLES
import math
import re
import warnings

class MyTestCase(unittest.TestCase):
    if not hasattr(unittest.TestCase, 'skipTest'):
        def skipTest(self, reason):
            return reason


class BaseTestCase(MyTestCase):
    TEXT_PASS = 'Module(body=[Pass()])'
    TEXT_PASS_PASS = 'Module(body=[Pass(), Pass()])'

    def _text_module(self, expr):
        return 'Module(body=[Expr(value=%s)])' % expr

    def _text_num(self, value):
        return 'Num(n=%r)' % value

    def text_num(self, value):
        return self._text_module(self._text_num(value))

    def _text_name(self, value):
        return 'Name(id=%r, ctx=Load())' % value

    def text_name(self, value):
        return self._text_module(self._text_name(value))

    def _text_bool(self, value):
        name = "True" if value else "False"
        return self._text_name(name)

    def text_bool(self, value):
        return self._text_module(self._text_bool(value))

    if PYTHON3:
        def _text_bytes(self, value):
            return "Bytes(s=%r)" % value
        def _text_unicode(self, value):
            return "Str(s=%r)" % value
    else:
        def _text_bytes(self, value):
            return "Str(s=%r)" % value
        def _text_unicode(self, value):
            return "Str(s=%r)" % value

    def text_bytes(self, value):
        return self._text_module(self._text_bytes(value))

    def text_unicode(self, value):
        return self._text_module(self._text_unicode(value))

    def text_native_str(self, value):
        return self._text_module("Str(s=%r)" % value)

    def _text_item(self, value):
        if value is None:
            return self._text_name('None')
        elif isinstance(value, (int, float, complex)):
            return self._text_num(value)
        elif isinstance(value, BYTES_TYPE):
            return self._text_bytes(value)
        elif isinstance(value, UNICODE_TYPE):
            return self._text_unicode(value)
        else:
            raise NotImplementedError("Unknown type: %s" % type(value))

    def _text_tuple(self, *elts):
        elts = (self._text_item(elt) for elt in elts)
        elts = '[%s]' % ', '.join(elts)
        return 'Tuple(elts=%s, ctx=Load())' % elts

    def text_tuple(self, *elts):
        return self._text_module(self._text_tuple(*elts))

    def text_frozenset(self, *elts):
        elts = tuple(frozenset(elts))
        arg = self._text_tuple(*elts)
        func = self._text_name('frozenset')
        call = 'Call(func=%s, args=[%s], keywords=[], starargs=None, kwargs=None)' % (func, arg)
        return self._text_module(call)

    def text_ast(self, code):
        tree = parse_ast(code)
        return ast.dump(tree)

    def _optimize_ast(self, tree, config=None):
        optimizer = Optimizer(config)
        tree = optimizer.optimize(tree)
        if sys.version_info >= (2,6):
            # Ensure that the tree is compilable to bytecode
            compile_ast(tree)
        return tree, optimizer

    def check(self, code, expected, config=None):
        tree = parse_ast(code)
        tree, optimizer = self._optimize_ast(tree, config)
        text = ast.dump(tree)
        self.assertEqual(text, expected)

    def check_not_optimized(self, code, config=None):
        old_tree = parse_ast(code)
        old = ast.dump(old_tree)
        old = re.sub(r"UnaryOp\(op=USub\(\), operand=Num\(n=([0-9]+(?:\.[0-9]*)?)\)\)", lambda regs: "Num(n=-%s)" % regs.group(1), old)
        new_tree, optimizer = self._optimize_ast(old_tree, config)
        new = ast.dump(new_tree)
        self.assertEqual(new, old)

    def check_qualnames(self, code, expected):
        tree = parse_ast(code)
        tree, optimizer = self._optimize_ast(tree)
        self.assertEqual(optimizer.namespace.qualnames, expected)

    def check_values(self, code, expected):
        tree = parse_ast(code)
        tree, optimizer = self._optimize_ast(tree)
        self.assertEqual(optimizer.namespace.values, expected)


class TestOptimizer(BaseTestCase):
    def test_None(self):
        self.check('(None,) * 3', self.text_tuple(None, None, None))

        config = Config()
        config.add_constant('x', None)
        self.check('x', self.text_name('None'), config)

    def test_BinOp_int(self):
        self.check('1 + 2', self.text_num(3))
        self.check('10 - 1', self.text_num(9))
        self.check('2 * 3', self.text_num(6))
        self.check('21 // 5', self.text_num(4))
        self.check_not_optimized('21 // 0')
        self.check('21 % 5', self.text_num(1))
        self.check_not_optimized('21 % 0')
        self.check('10 ** 3', self.text_num(1000))
        self.check('-25 ** 0', self.text_num(-1))
        self.check('-2 ** -1', self.text_num(-0.5))
        self.check('1 << 5', self.text_num(32))
        self.check('32 >> 3', self.text_num(4))
        self.check_not_optimized('3 << -1')
        self.check('5 & 3', self.text_num(1))
        self.check('5 | 3', self.text_num(7))
        self.check('5 ^ 3', self.text_num(6))
        if PYTHON3:
            self.check('21 / 5', self.text_num(4.2))
        else:
            self.check_not_optimized('21 / 5')
        self.check_not_optimized('21 / 0')

        # complex operations
        self.check('1 + 2 * 3', self.text_num(7))

        # simple optimizations
        self.check_not_optimized('0 + x')
        self.check_not_optimized('0.0 + x')
        self.check_not_optimized('x + 0')
        self.check_not_optimized('x + 0.0')
        self.check_not_optimized('1 * x')
        self.check_not_optimized('1.0 * x')
        self.check_not_optimized('x * 1')
        self.check_not_optimized('x * 1.0')

        # test min_int / max_int limits
        config = Config()
        config.min_int = -10
        config.max_int = 10
        self.check('1 + 2', self.text_num(3), config)
        self.check_not_optimized('4 * 5', config)
        self.check_not_optimized('2 - 40', config)

    def test_BinOp_float(self):
        self.check('1.0 + 2', self.text_num(3.0))
        self.check('10 - 1.0', self.text_num(9.0))
        self.check('2.0 * 3.0', self.text_num(6.0))
        self.check('21.0 // 5.0', self.text_num(4.0))
        self.check('21.0 % 5.0', self.text_num(1.0))
        self.check('10.0 ** 3', self.text_num(1000.0))
        self.check_not_optimized('(-25) ** 0.5')

        self.check_not_optimized('1.0 << 5')
        self.check_not_optimized('32.0 >> 3')
        self.check_not_optimized('5.0 & 3')
        self.check_not_optimized('5.0 | 3')
        self.check_not_optimized('5.0 ^ 3')

    def test_BinOp_complex(self):
        self.check('1j + 2', self.text_num(1j + 2.0))
        self.check('10 - 1j', self.text_num(10.0 - 1j))
        self.check('2j * 3.0', self.text_num(6j))
        self.check('21.0 / 5.0', self.text_num(4.2))
        if PYTHON3:
            self.check_not_optimized('21j // 5j')
            self.check_not_optimized('21j % 5j')
        elif sys.version_info >= (2, 6):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", DeprecationWarning)
                self.check('21j // 5j', self.text_num(4+0j))
                self.check('21j % 5j', self.text_num(1j))
        else:
            # Python <= 2.5
            self.check('21j // 5j', self.text_num(4+0j))
            self.check('21j % 5j', self.text_num(1j))
        self.check_not_optimized('10j ** 3')
        self.check_not_optimized('2 ** 1j')
        self.check_not_optimized('1.0 << 5')
        self.check_not_optimized('32.0 >> 3')
        self.check_not_optimized('5.0 & 3')
        self.check_not_optimized('5.0 | 3')
        self.check_not_optimized('5.0 ^ 3')

    def test_BinOp_numbers(self):
        self.check('1 + 2.0', self.text_num(3.0))
        self.check('1.0 + 2', self.text_num(3.0))
        self.check('1.0 + 2 * 3.0', self.text_num(7.0))

    def test_BinOp_string(self):
        self.check('%r * 3' % b("a"), self.text_bytes(b('aaa')))
        self.check('%r * 3' % (u('a'),), self.text_unicode(u('aaa')))
        self.check('%r %% 123' % u("x=%s"), self.text_unicode(u('x=123')))
        self.check('%r %% (2, 3)' % u("x=%s y=%s"), self.text_unicode(u('x=2 y=3')))
        if not PYTHON3:
            self.check('%r %% 123' % b("x=%s"),
                       self.text_bytes(b('x=123')))
            self.check('%r %% (2, 3)' % b("x=%s y=%s"),
                       self.text_bytes(b('x=2 y=3')))
            self.check('%r %% 4' % b("x\xe9=%s"),
                       self.text_bytes(b('x\xe9=4')))
            self.check_not_optimized('%r %% %r' % (b("x\xe9=%s"), u('abc')))
        else:
            self.check_not_optimized('%r %% 123' % b("x=%s"))
            self.check_not_optimized('%r %% (2, 3)' % b("x=%s y=%s"))
        self.check_not_optimized('"a" - "b"')
        self.check_not_optimized('"a" * "b"')
        self.check_not_optimized('"a" * 2.0')
        self.check_not_optimized('%r %% x' % b("x=%s"))
        self.check_not_optimized('%r %% 2' % b("x=%s y=%s"))
        self.check_not_optimized('%r %% x' % u("x=%s"))
        self.check_not_optimized('%r %% 2' % u("x=%s y=%s"))

        config = Config()
        config.max_string_length = 5
        self.check('"a"*5', self.text_unicode('aaaaa'), config)
        self.check_not_optimized('"a"*6', config)

    def test_BinOp_tuple(self):
        self.check('(1,) + (2,)', self.text_tuple(1, 2))
        self.check('(1, 2) * 2', self.text_tuple(1, 2, 1, 2))
        self.check('2 * (1, 2)', self.text_tuple(1, 2, 1, 2))

        config = Config()
        config.max_tuple_length = 3
        self.check('(1,) * 3', self.text_tuple(1, 1, 1), config)
        self.check_not_optimized('(1,) * 4', config)

    def test_BinOp_frozenset(self):
        self.check('frozenset("ab") | frozenset("bc")', self.text_frozenset("a", "b", "c"))
        self.check('frozenset("ab") ^ frozenset("bc")', self.text_frozenset("a", "c"))
        self.check('frozenset("abc") - frozenset("c")', self.text_frozenset("a", "b"))
        self.check('frozenset("abc") & frozenset("bc")', self.text_frozenset("b", "c"))

    def test_BoolOp(self):
        # a and b constant
        self.check('True and True', self.text_bool(True))
        self.check('False or True', self.text_bool(True))
        self.check('13 or 14', self.text_num(13))

        # a constant
        self.check('True and x', self.text_name('x'))
        self.check('False and x', self.text_bool(False))
        self.check('True or x', self.text_bool(True))
        self.check('False or x', self.text_name('x'))

        # b constant
        self.check_not_optimized('x and True')
        self.check_not_optimized('x and False')
        self.check_not_optimized('x or True')
        self.check_not_optimized('x or False')

        # others
        self.check('4 and 5 and x and 6', self.text_ast('x and 6'))
        self.check('4 and 5 and ~6', self.text_num(-7))

    def test_UnaryOp(self):
        self.check('not True', self.text_bool(False))
        self.check('not 3', self.text_bool(False))
        self.check('not False', self.text_bool(True))
        self.check('not 0', self.text_bool(True))
        self.check_not_optimized('not x')

        self.check('~3', self.text_num(-4))
        self.check_not_optimized('~x')

        self.check('-(0.0)', self.text_num(-0.0))
        self.check('-(2)', self.text_num(-2))
        self.check('-(-(3))', self.text_num(3))
        self.check('+(0.0)', self.text_num(0.0))
        self.check('+(2)', self.text_num(2))
        self.check('+(-(3))', self.text_num(-3))

        config = Config()
        config.min_int = -10
        config.max_int = 10
        self.check('~9', self.text_num(-10), config)
        self.check_not_optimized('~10', config)

    def test_Subscript(self):
        self.check('%r[:3]' % b("abcdef"), self.text_bytes(b("abc")))
        if PYTHON3:
            self.check('%r[1]' % b("abcdef"), self.text_num(ord(b("b"))))
        else:
            self.check('%r[1]' % b("abcdef"), self.text_unicode(b("b")))
        self.check('"abcdef"[:3]', self.text_unicode("abc"))
        self.check('"abcdef"[2:4]', self.text_unicode("cd"))
        self.check('"abcdef"[2::-1]', self.text_unicode("cba"))
        self.check('"abcdef"[2:0:-1]', self.text_unicode("cb"))
        self.check('"abcdef"[1]', self.text_unicode("b"))
        if sys.version_info < (3,3):
            self.check_not_optimized('%r[4]' % u("abc\U0010ffff-"))
            self.check_not_optimized('%r[-1]' % u("abc\U0010ffff-"))
            self.check_not_optimized('%r[::-1]' % u("abc\U0010ffff-"))
            self.check_not_optimized('%r[3::-1]' % u("abc\U0010ffff-"))
            self.check_not_optimized('%r[3:0:-1]' % u("abc\U0010ffff-"))
        else:
            self.check('%r[3]' % u("abc\U0010ffff-"), self.text_unicode(u("\U0010ffff")))
            self.check('%r[-1]' % u("abc\U0010ffff-"), self.text_unicode("-"))

        self.check('"abc"[2]', self.text_native_str('c'))
        self.check_not_optimized('"abc"[3]')
        self.check('"abc"[-3]', self.text_native_str('a'))
        self.check_not_optimized('"abc"[-4]')

    def test_If(self):
        # FIXME: issue with "def func(): if 0: yield; return"
        # http://tomlee.co/2008/04/the-internals-of-python-generator-functions-in-the-ast/
        self.check('if 0: print("log")', self.TEXT_PASS)
        self.check('if 1: 3', self.text_num(3))
        self.check('if 2**2**2 - 16: 3', self.TEXT_PASS)
        # FIXME: Why do we get pass; pass ?
        self.check('if 0: pass\nelif 0: pass\npass', self.TEXT_PASS_PASS)

    def test_While(self):
        # FIXME: issue with "def func(): while 0: yield; return"
        # http://tomlee.co/2008/04/the-internals-of-python-generator-functions-in-the-ast/
        self.check('while 0: print("log")', self.TEXT_PASS)

        code = '\n'.join((
            'while 0:',
            '    print("log")',
            'else:',
            '    print("abc")',
        ))
        self.check(code, self.text_ast('print("abc")'))

    def test_Compare(self):
        self.check('"abc" == "abc"', self.text_bool(True))
        self.check('1 != 2', self.text_bool(True))
        self.check('2 < 2', self.text_bool(False))
        self.check('2 <= 2', self.text_bool(True))
        self.check('2 > 2', self.text_bool(False))
        self.check('2 >= 2', self.text_bool(True))
        self.check('"xxx" in "abcdef"', self.text_bool(False))
        self.check('"xxx" not in "abcdef"', self.text_bool(True))
        self.check('"bcd" in "abcdef"', self.text_bool(True))
        self.check('"bcd" not in "abcdef"', self.text_bool(False))
        self.check('0 in (1, 2, 3)', self.text_bool(False))
        self.check('2 in (1, 2, 3)', self.text_bool(True))
        self.check('"x" in frozenset("abc")', self.text_bool(False))
        self.check('"x" not in frozenset("abc")', self.text_bool(True))
        self.check('"b" in frozenset("abc")', self.text_bool(True))
        self.check('frozenset("abc") <= frozenset("abc")', self.text_bool(True))
        self.check('frozenset("abc") < frozenset("abc")', self.text_bool(False))
        self.check('True is True', self.text_bool(True))
        self.check('False is False', self.text_bool(True))
        self.check('None is None', self.text_bool(True))
        self.check('True is None', self.text_bool(False))
        self.check_not_optimized('"" is ""')
        self.check_not_optimized('1 is 1')
        self.check('True is not True', self.text_bool(False))
        self.check('True is not None', self.text_bool(True))
        self.check_not_optimized('"" is not ""')
        self.check_not_optimized('1 is not 1')
        # may raise a BytesWarning exception
        self.check_not_optimized('%r == %r' % (b("abc"), u("abc")))
        self.check_not_optimized('%r in %r' % (b("abc"), u("abc")))
        self.check_not_optimized('%r not in %r' % (b("abc"), u("abc")))
        # raise a TypeError
        self.check_not_optimized('(1,) < frozenset()')

    def test_Name(self):
        config = Config()
        config.add_constant('debug', 0)
        self.check('debug', self.text_num(0), config)

    def test_Assign(self):
        if ENABLE_VARIABLES:
            self.check_values('x=1', {'x': 1})
            self.check_values('x,y=1,2', {'x': 1, 'y': 2})

            self.check_values('x=1; x=2', {'x': 2})
            self.check_values('x,y=1,2; x,y=y,x', {'x': 2, 'y': 1})

            self.check('x=math.e; print(x)',
                       self.text_ast('x=%r; print(%r)' % (math.e, str(math.e))))

            # disable optimization on mutable types
            self.check_values('x=[]', {})
            self.check_values('x=[]; x=2', {})
            self.check_values('x=(1, 2); x[:1] = (2,)', {})
            self.check_values('x=1; x += 1', {})
            self.check_values('x=1; del x', {})

        config = Config()
        config.add_constant('x', 42)
        self.check('x', self.text_num(42), config)
        self.check_not_optimized('x=1', config)

    def test_Print(self):
        if PYTHON3:
            self.check('print(1.5)', self.text_ast('print(%r)' % ('1.5',)))
            self.check('print(1, 2)', self.text_ast('print(%r)' % ('1 2',)))
            self.check_not_optimized('print(1, 2, sep="-")')
            self.check_not_optimized('print(1, 2, end="")')
        else:
            self.check('print 1.5', self.text_ast('print %r' % ('1.5',)))
            self.check('print 1, 2', self.text_ast('print %r' % ('1 2',)))
            self.check('print 1, 2,', self.text_ast('print %r,' % ('1 2',)))
            self.check('print >>f, "x =", 2', self.text_ast('print >> f, %r' % ('x = 2',)))

    def test_max_tuple_length(self):
        config = Config()
        config.max_tuple_length = 3
        self.check('len((1,2,3))', self.text_num(3), config)
        self.check_not_optimized('len((1,2,3,4))', config)
        self.check('len(frozenset("abc"))', self.text_num(3), config)
        self.check_not_optimized('len(frozenset("abcd"))', config)

class TestBuiltins(BaseTestCase):
    def test_Call_builtins(self):
        self.check('len("abc")', self.text_num(3))
        self.check('len("abcdef"[:2])', self.text_num(2))
        self.check('len((1,2,3))', self.text_num(3))
        self.check('len((1,2,3,4,5)[:2])', self.text_num(2))
        self.check('abs(-5)', self.text_num(5))
        self.check_not_optimized('abs("abc")')
        if sys.version_info >= (2,6):
            self.check('bin(5)', self.text_unicode('0b101'))
            self.check_not_optimized('bin(3.5)')
        self.check('bool(0)', self.text_bool(False))
        self.check('bool(3)', self.text_bool(True))
        self.check('bool("")', self.text_bool(False))
        self.check('bool("xxx")', self.text_bool(True))
        self.check('chr(65)', self.text_unicode('A'))
        self.check('divmod(10, 3)', self.text_ast('(3, 1)'))
        self.check('divmod(10.5, 3.0)', self.text_ast('(3.0, 1.5)'))
        self.check('float(6)', self.text_num(6.0))
        self.check('float(6.9)', self.text_num(6.9))
        self.check_not_optimized('float(1j)')
        self.check('int(6)', self.text_num(6))
        self.check('int(6.9)', self.text_num(6))
        self.check_not_optimized('int(1j)')
        self.check_not_optimized('oct(3.5)')

        self.check('ord(%r)' % b("A"), self.text_num(65))
        self.check('ord(%r)' % u("A"), self.text_num(65))
        self.check_not_optimized('ord("abc")')
        if sys.version_info >= (3,3):
            self.check('ord(%r)' % u("\U0010ffff"), self.text_num(0x10ffff))
        else:
            self.check_not_optimized('ord(%r)' % u("\U0010ffff"))
        config = Config('pythonenv')
        if sys.maxunicode == 0x10ffff:
            self.check('ord(%r)' % u("\U0010ffff"), self.text_num(0x10ffff), config)
        else:
            self.check_not_optimized('ord(%r)' % u("\U0010ffff"), config)

        self.check("min('abc')", self.text_native_str('a'))
        self.check("min((1, 2, 3))", self.text_num(1))
        self.check("min(frozenset((1, 2, 3)))", self.text_num(1))
        self.check_not_optimized("min(1)")
        self.check("max('abc')", self.text_native_str('c'))
        self.check("max((1, 2, 3))", self.text_num(3))
        self.check("max(frozenset((1, 2, 3)))", self.text_num(3))
        self.check_not_optimized("max(1)")

        self.check('pow(2, 3)', self.text_num(8))
        self.check('pow(2, 123, 4567)', self.text_num(197))
        self.check('pow(10.0, 3)', self.text_num(1000.0))
        self.check_not_optimized('pow(-25.0, 0.5)')
        self.check_not_optimized('pow(10.0, 3, 7)')
        self.check('repr(1)', self.text_unicode('1'))
        self.check("repr('abc')", self.text_unicode("'abc'"))
        self.check('str(1)', self.text_unicode('1'))
        self.check("str('abc')", self.text_unicode('abc'))
        self.check("sum((1, 2, 3))", self.text_num(6))
        self.check("sum(frozenset((1, 2, 3)))", self.text_num(6))
        self.check("sum((1, 2, 3), 0)", self.text_num(6))
        self.check_not_optimized("sum(((1,), 2))")
        self.check_not_optimized("sum((1, 2), 'a')")
        self.check("tuple()", self.text_tuple())
        self.check("tuple('abc')", self.text_tuple('a', 'b', 'c'))
        self.check("tuple((1, 2, 3.0))", self.text_tuple(1, 2, 3.0))

        if PYTHON3:
            self.check('oct(5)', self.text_unicode('0o5'))
            self.check("round(5)", self.text_num(5))
            self.check("round(5.5)", self.text_num(6))
            self.check("round(5, 1)", self.text_num(5))
            self.check("round(5.567, 1)", self.text_num(5.6))
        else:
            self.check('oct(5)', self.text_unicode('05'))
            self.check('long(6)', self.text_num(long(6)))
            self.check('long(6.9)', self.text_num(long(6)))
            self.check_not_optimized('long(1j)')
            self.check('unichr(65)', self.text_unicode(u('A')))
            self.check('repr(1L)', self.text_unicode('1L'))
            self.check('str(1L)', self.text_unicode('1'))
            self.check('unicode(1)', self.text_unicode(u('1')))
            self.check('unicode(1L)', self.text_unicode(u('1')))
            self.check("unicode('abc')", self.text_unicode(u('abc')))
            self.check_not_optimized("unicode('abc\xff')")
            self.check("round(5)", self.text_num(5.0))
            self.check("round(5.5)", self.text_num(6.0))
            self.check("round(5, 1)", self.text_num(5.0))
            self.check("round(5.567, 1)", self.text_num(5.6))

        config = Config()
        config.min_int = -10
        config.max_int = 10
        config.max_string_length = 5
        self.check('pow(2, 3)', self.text_num(8), config)
        self.check_not_optimized('pow(2, 10)', config)
        self.check('pow(2, 10, 7)', self.text_num(2), config)
        self.check_not_optimized('pow(2, 10, 42)', config)
        self.check('repr(12345)', self.text_unicode('12345'), config)
        self.check_not_optimized('repr(1234567)', config)
        self.check('str(12345)', self.text_unicode('12345'), config)
        self.check_not_optimized('str(1234567)', config)

        config = Config()
        config.max_size = 3
        self.check('len((1,2,3))', self.text_num(3), config)
        self.check_not_optimized('len((1,2,3,4))', config)

        self.check('float("6")', self.text_num(6.0))
        self.check('float("6.9")', self.text_num(6.9))
        self.check('float("inf")', self.text_num(float("inf")))
        self.check('float("-inf")', self.text_num(float("-inf")))
        self.check('float("nan")', self.text_num(float("nan")))
        self.check('float("-0.0")', self.text_num(float("-0.0")))
        self.check_not_optimized('float("abc")')

        config = Config()
        config.max_tuple_length = 3
        self.check('tuple("abc")', self.text_ast('("a", "b", "c")'), config)
        self.check_not_optimized('tuple("abcd")', config)

        if sys.version_info >= (3,3):
            self.check('chr(0xffff)', self.text_unicode(chr(0xffff)))
            self.check('chr(0x10ffff)', self.text_unicode(chr(0x10ffff)))
            self.check_not_optimized('chr(0x110000)')
        elif PYTHON3:
            self.check('chr(0xffff)', self.text_unicode(chr(0xffff)))
            self.check_not_optimized('chr(0x10ffff)')
            config = Config('pythonenv')
            self.check('chr(0x10ffff)', self.text_unicode(chr(0x10ffff)), config)
            self.check_not_optimized('chr(0x110000)')
        else:
            self.check_not_optimized('chr(0x100)')
            self.check('unichr(0xffff)', self.text_unicode(unichr(0xffff)))
            self.check_not_optimized('unichr(0x10ffff)')
            if sys.maxunicode > 0xffff:
                config = Config('pythonenv')
                self.check('unichr(0x10ffff)', self.text_unicode(unichr(0x10ffff)), config)
            self.check_not_optimized('unichr(0x110000)')

    def test_str_methods(self):
        self.check('%r.decode("ascii")' % b("abc"), self.text_unicode(u('abc')))
        self.check('%r.decode("ascii", "replace")' % b("ab\xff"), self.text_unicode(u('ab\ufffd')))
        self.check('%r.encode("ASCII")' % u('abc'), self.text_bytes(b('abc')))
        self.check('%r.encode("ASCII", "ignore")' % u('ab\xff'), self.text_bytes(b('ab')))
        self.check('%r.encode("ASCII", "backslashreplace")' % u('ab\xff'), self.text_bytes(b('ab\\xff')))
        self.check('%r.encode("latin1")' % u('abc'), self.text_bytes(b('abc')))
        self.check('%r.count(%r)' % (b('abab'), b('a')), self.text_num(2))
        self.check('%r.find(%r)' % (b('abc'), b('b')), self.text_num(1))
        self.check('%r.index(%r)' % (b('abc'), b('b')), self.text_num(1))
        self.check('%r.startswith(%r)' % (b('abc'), b('ab')), self.text_bool(True))
        self.check('%r.endswith(%r)' % (b('abc'), b('xx')), self.text_bool(False))
        self.check('%r.count(%r)' % (u('abab'), u('a')), self.text_num(2))
        self.check('%r.find(%r)' % (u('abc'), u('b')), self.text_num(1))
        self.check('%r.index(%r)' % (u('abc'), u('b')), self.text_num(1))
        self.check('%r.startswith(%r)' % (u('abc'), u('x')), self.text_bool(False))
        self.check('%r.endswith(%r)' % (u('abc'), u('bc')), self.text_bool(True))
        self.check_not_optimized('%r.index(%r)' % (u('abc'), u('xx')))

        if sys.version_info < (3,3):
            self.check_not_optimized('%r.encode("utf-8")' % u('\U0010ffff'))
        else:
            self.check(
                '%r.encode("utf-8")' % u('\U0010ffff'),
                self.text_bytes(b('\xf4\x8f\xbf\xbf')))

    def test_tuple_methods(self):
        if sys.version_info >= (2, 6):
            self.check('(2,0,2,2).count(2)', self.text_num(3))
            self.check('(2,0,2,2).index(0)', self.text_num(1))
            self.check_not_optimized('(2,0,2,2).index(9)')

    def test_int_methods(self):
        if sys.version_info >= (2, 7):
            self.check('(12345).bit_length()', self.text_num(14))

    def test_float_methods(self):
        if sys.version_info >= (2,6):
            self.check('(1.0).is_integer()', self.text_bool(True))
            self.check('(1.3).is_integer()', self.text_bool(False))
            self.check('(1.5).as_integer_ratio()', self.text_tuple(3, 2))
            self.check('(1.5).hex()', self.text_native_str('0x1.8000000000000p+0'))
            self.check('float.fromhex("0x1.8000000000000p+0")', self.text_num(1.5))


class TestFeatures(BaseTestCase):
    def test_platform(self):
        import errno
        import os
        import platform
        import socket

        self.check_not_optimized('os.O_RDONLY')
        self.check_not_optimized('errno.ENOENT')
        self.check_not_optimized('sys.platform')

        config = Config('platform')
        self.check('os.O_RDONLY', self.text_num(os.O_RDONLY), config)
        self.check('errno.ENOENT', self.text_num(errno.ENOENT), config)
        self.check('socket.AF_INET', self.text_num(socket.AF_INET), config)
        self.check('socket.SOCK_STREAM', self.text_num(socket.SOCK_STREAM), config)
        self.check('sys.platform', self.text_native_str(sys.platform), config)
        self.check('platform.machine()', self.text_native_str(platform.machine()), config)
        self.check('platform.release()', self.text_native_str(platform.release()), config)
        self.check('platform.system()', self.text_native_str(platform.system()), config)
        self.check('socket.ntohl(12)', self.text_num(socket.ntohl(12)), config)
        self.check('socket.ntohs(12)', self.text_num(socket.ntohs(12)), config)
        self.check('socket.htonl(12)', self.text_num(socket.htonl(12)), config)
        self.check('socket.htons(12)', self.text_num(socket.htons(12)), config)

    def test_pythonenv(self):
        import platform

        self.check_not_optimized('__debug__')
        self.check_not_optimized('sys.hexversion')
        self.check_not_optimized('sys.platform')
        self.check_not_optimized('platform.python_version')
        if not PYTHON3:
            self.check_not_optimized('21 / 5')
        if sys.version_info < (3,3):
            self.check_not_optimized('%r[-1]' % u("abc\U0010ffff-"))

        config = Config('pythonenv')
        self.check('__debug__', self.text_bool(__debug__), config)
        self.check('sys.hexversion', self.text_num(sys.hexversion), config)
        self.check('sys.version[:3]', self.text_native_str(sys.version[:3]), config)
        self.check('platform.python_version()', self.text_native_str(platform.python_version()), config)
        if sys.version_info >= (2, 6):
            self.check('platform.python_version_tuple()', self.text_tuple(*platform.python_version_tuple()), config)
        self.check('21 / 5', self.text_num(21 / 5), config)
        if sys.version_info < (3,3):
            if sys.maxunicode == 0x10ffff:
                self.check(
                    '%r[3]' % u("abc\U0010ffff-"),
                    self.text_unicode(u("\U0010ffff")),
                    config)
            else:
                self.check(
                    '%r[3]' % u("abc\U0010ffff-"),
                    self.text_unicode(u("\udbff")),
                    config)
                self.check(
                    '%r[4]' % u("abc\U0010ffff-"),
                    self.text_unicode(u("\udfff")),
                    config)
            self.check(
                '%r[-1]' % u("abc\U0010ffff-"),
                self.text_unicode(u("-")),
                config)

    def test_math(self):
        self.check('math.e', self.text_num(math.e))
        self.check('math.pi', self.text_num(math.pi))
        self.check('math.exp(1.0)', self.text_num(math.e))
        self.check('math.log(1.0)', self.text_num(0.0))
        self.check('math.log(math.e)', self.text_num(1.0))
        self.check('math.pow(2, 4)', self.text_num(16.0))
        self.check('math.pow(10.0, 3)', self.text_num(1000.0))
        self.check_not_optimized('math.pow(2j, 4)')
        self.check_not_optimized('math.pow(-25.0, 0.5)')

        self.check('math.sqrt(4.0)', self.text_num(2.0))
        self.check_not_optimized('math.sqrt(-1)')
        if PYTHON3:
            self.check('math.ceil(1.5)', self.text_num(2))
            self.check('math.floor(1.5)', self.text_num(1))
        else:
            self.check('math.ceil(1.5)', self.text_num(2.0))
            self.check('math.floor(1.5)', self.text_num(1.0))
        if sys.version_info >= (2, 6):
            self.check('math.trunc(4.5)', self.text_num(4))
            self.check('math.trunc(-1.5)', self.text_num(-1))

    def test_Import(self):
        self.check_qualnames('import math as M', {'M': 'math'})
        self.check('import math as M; x=M.sqrt(4.0)',
                   self.text_ast('import math as M; x=2.0'))
        code = '\n'.join((
            'try:',
            '    import StringIO as BUFFER',
            'except ImportError:',
            '    import cStringIO as BUFFER',
        ))
        self.check_qualnames(code, {})

    def test_ImportFrom(self):
        self.check_qualnames('from math import trunc as M', {'M': 'math.trunc'})
        code = '\n'.join((
            'try:',
            '    from StringIO import StringIO as BUFFER',
            'except ImportError:',
            '    from cStringIO import StringIO as BUFFER',
        ))
        self.check_qualnames(code, {})


class TestNamespace(MyTestCase):
    def test_module_aliases(self):
        ns = Namespace()
        ns.add_alias('math', 'xxx')
        self.assertEqual(ns.get_qualname('xxx'), 'math')

        ns2 = Namespace()
        ns2.add_alias('math.trunc', 'trunc')
        self.assertEqual(ns2.get_qualname('trunc'), 'math.trunc')

        ns3 = Namespace()
        ns3.add_alias('os', 'O')
        ns3.add_alias('os.path', 'P')
        ns3.add_alias('os.path.exists', 'E')
        self.assertEqual(ns3.get_qualname('O'), 'os')
        self.assertEqual(ns3.get_qualname('P'), 'os.path')
        self.assertEqual(ns3.get_qualname('E'), 'os.path.exists')


def main():
    unittest.main(module=__name__)

if __name__ == "__main__":
    main()

